#!/usr/bin/perl -w

###########################################################################
# This script takes as input a graph and outputs a Reflexion puzzle that
# has a solution if and only if the graph possesses a Hamiltonian cycle.
# The graph must be given in a format inspired by the DIMACS format: the
# first line is either "p dir" for a directed graph, or "p undir" for an
# undirected graph. After that every line contains two natural numbers
# (e.g., "1 2") indicating an edge from node 1 to node 2. If the graph
# contains n nodes, its identifiers should be 0..n-1. Commentary lines
# start with c, as in DIMACS.
###########################################################################

$PROG = "hamilton";

########### process command-line options ###########

sub usage {
	print <<EOT;

$PROG -- generates a Reflexion puzzle that searches for a Hamiltonian cycle

usage: $PROG [<filename>]
		<filename> : contains directed graph
		if no filename is given, reads from stdin

	Options:
		-h: print this help message
		-q: quiet (reduce messages)

EOT
		exit 2;
}

$options = {
	'h' => \$opt_help,
	'q' => \$opt_quiet
};

for $v (values $options) { $$v = 0; }
$fn = undef;
for $arg (@ARGV) {
	if (substr($arg,0,1) eq "-") {
		for $a (split(//,substr($arg,1))) {
			die "unknown option -$a" unless defined($options->{$a});
			${$options->{$a}} = 1;
		}
	} else {
		usage() if defined($fn);
		$fn = $arg;
	}
}

usage() if $opt_help;

print STDERR "$PROG -- (invoke with -h for help)\n" unless $opt_quiet;
if (defined($fn)) {
	die "cannot read file $fn" unless -r $fn;
	open STDIN,$fn;
	print STDERR "reading graph from file $fn\n" unless $opt_quiet;
} else {
        print STDERR "reading graph from standard input\n" unless $opt_quiet;
}

########### read graph description ###########

$n = 0;	# nodes
$m = 0; # edges
@edges = ();
$undir = "";

while (<STDIN>) {
	tr/\015\012//d;
	if (/^p ((un)?)dir/) { $undir = $1; next; }
	next if /^c/;
	die "bad file format" unless /^\s*\d+\s*\d+\s*$/;
	($a,$b) = split;
	$n = $a if $a > $n;
	$n = $b if $b > $n;
	push @edges,"$a $b";
	push @edges,"$b $a" if $undir;
}

# catch annoying corner cases
die "two nodes, one edge => no solution" if $undir && $n == 1 && $#edges == 1;
die "one node, no edge => no solution" if $n == 0 && $#edges == -1;

print STDERR "generating puzzle for ${undir}directed graph...\n\n"
	unless $opt_quiet;

########### compute some properties of the graph ###########

$n++;
$m = $#edges+1;
$maxin = 0;
$maxout = 0;
for ($i = 0; $i < $n; $i++) { $od[$i] = $id[$i] = 0; }
for $e (@edges) {
	($a,$b) = split(' ',$e);
	$od[$a]++;
	$id[$b]++;
	$maxout = $od[$a] if $od[$a] > $maxout;
	$maxin = $id[$b] if $id[$b] > $maxin;
}

########### generate the puzzle ###########

$width = 4 + 3*$n + $m + $maxout;
$height = 1 + 3*$n + $m + $maxin;

# puzzle with outer walls
@puzzle = ();
for ($i = 0; $i < $height; $i++) {
	my @line = ('%',(' ')x($width-2),'%',"\n");
	push @puzzle,\@line;
}
put(0,0,'%'x$width);
put(0,$height-1,'%'x$width);

# start/end and diamonds
put($n+$maxout+1,0,'S%E','M%M');
put($n+$maxout+2,2,('D')x($n-1));

# upper left mirrors
for ($i = 1; $i <= $n; $i++) { put($i,$i,'M'); }

# middle box
for ($i = 0; $i < $maxin; $i++) { put($n+1,$n+1+$i,'%'x($maxout+3)); }

# choice among outgoing edges
$row = $n+$maxin+1;
$col = $n+1;
for ($i = 0; $i < $n; $i++) {
	$opos[$i] = $row;
	$opos[$i] = $row+1 if $od[$i] == 1;
	put($i+1,$row+1,"M") unless $od[$i] == 0;
	put($col,$row+$od[$i]+1,'%'x$maxout);
	if ($od[$i] >= 2) {
		for ($j = 0; $j < $od[$i]-1; $j++) {
			put($col+$j,$row+$j,'M','M','M');
		}
		put($col+$j,$row+$j,'M','M');
	}
	$row += $od[$i]+2;
}

# merging incoming edges
$row = $n+1;
$col = $n+$maxout+4;
for ($i = 0; $i < $n; $i++) {
	$ipos[$i] = $col;
	$ipos[$i] = $col+1 if $id[$i] == 1;
	put($col+1,$i+1,"M") unless $id[$i] == 0;
	put($col+$id[$i]+1,$row,('%')x$maxin);
	if ($id[$i] >= 2) {
		for ($j = 0; $j < $id[$i]-1; $j++) {
			put($col+$j,$row+$j,'MMM');
		}
		put($col+$j,$row+$j,'MM');
	}
	$col += $id[$i]+2;
}

# fill lower right area with one mirror per edge
for $e (@edges) {
	($a,$b) = split(' ',$e);
	put($ipos[$b]++,$opos[$a]++,'M');
}

########### print puzzle to stdout ###########

for $line (@puzzle) { print join("",@$line); }
exit 0;

# auxilary procedure

sub put {
	my ($x,$y,@l) = @_;
	for my $s (@l) {
		my $line = $puzzle[$y++];
		my @nc = split(//,$s);
		splice @$line,$x,$#nc+1,@nc;
	}
}

