#!/usr/bin/perl -w

###########################################################################
# This script takes as input a graph and outputs a Reflexion puzzle that
# has a solution if and only if the graph has a vertex cover of size k.
# The graph must be in the same format as for hamilton-reflex-pl, see there.
# Directed and undirected graphs are treated identically.
# Additionally, k must be specified on the command line (default is 1).
###########################################################################

$PROG = "vertexcover";

########### process command-line options ###########

sub usage {
	print <<EOT;

$PROG -- generates a Reflexion puzzle that searches for a k-vertex cover

usage: $PROG [<filename>]
		<filename> : contains directed graph
		if no filename is given, reads from stdin

	Options:
		-d: generate extra diamonds (helps the solver sometimes)
		-h: print this help message
		-k=<int>: set value of k (default 1)
		-q: quiet (reduce messages)

EOT
		exit 2;
}

$options = {
	'd' => \$opt_diam,
	'h' => \$opt_help,
	'q' => \$opt_quiet
};

for $v (values $options) { $$v = 0; }
$fn = undef;
$k = 1;
for $arg (@ARGV) {
	if (substr($arg,0,1) eq "-") {
		for ($i = 1; $i < length($arg); $i++) {
			$a = substr($arg,$i,1);
			if ($a eq 'k') {
				$b = substr($arg,$i+1);
				$b =~ s/^(=?)(\d*)//;
				$k = $2;
				die "no value for k given" if $k eq "";
				$k = int($k);
				$arg = substr($arg,0,$i+1).$b;
				next;
			}
			die "unknown option -$a" unless defined($options->{$a});
			${$options->{$a}} = 1;
		}
	} else {
		usage() if defined($fn);
		$fn = $arg;
	}
}

usage() if $opt_help;
die "k must be positive" if $k == 0;

print STDERR "$PROG -- (invoke with -h for help)\n" unless $opt_quiet;
if (defined($fn)) {
	die "cannot read file $fn" unless -r $fn;
	open STDIN,$fn;
	print STDERR "reading graph from file $fn\n" unless $opt_quiet;
} else {
        print STDERR "reading graph from standard input\n" unless $opt_quiet;
}

########### read graph description ###########

$n = 0;	# nodes
@edges = ();
@loops = ();

while (<STDIN>) {
	tr/\015\012//d;
	next if (/^p ((un)?)dir/);
	next if /^c/;
	die "bad file format" unless /^\s*\d+\s*\d+\s*$/;
	($a,$b) = split;
	$n = $a if $a > $n;
	$n = $b if $b > $n;
	if ($a == $b) { push @loops,$a; next; }
	push @edges,(($a < $b)? "$a $b" : "$b $a");
}

$n++;
die "instance is trivially solvable (k >= n)" if $n <= $k;
print STDERR "generating puzzle for k=$k...\n\n" unless $opt_quiet;

########### generate the puzzle ###########

# compute some important values
for ($i = 0; $i < $n-$k; $i++) { $outrow[$i] = 1+2*$i; }
for ($i = $n-$k; $i < $n; $i++) { $outrow[$i] = 2+4*$i-2*($n-$k); }
$half = 2*$n+2*$k-1;
$vpos = ($n-$k)*2+$k+1;
$width = $vpos+$n+2;
$height = 2*$half+1;

# puzzle with outer walls
@puzzle = ();
$baserow = 0;
for ($i = 0; $i < $height; $i++) {
	my @line = ('%',(' ')x($width-2),'%',"\n");
	push @puzzle,\@line;
}

put(0,0,'%'x$width);
put(0,$height-1,'%'x$width);

# start, restart, end
put(0,2,'S');
$dstr = $opt_diam? 'MD' : 'M';
for ($i = 1; $i < $k; $i++) { put($i,4*$i+2,$dstr); }
for ($i = 1; $i < $k; $i++) { put($i,4*$i-2+$half,$dstr); }
put(0,4*$k+$half-2,'E');

# merge/choice arrays
$pos = $k+1;
for ($i = 0; $i < $k-1; $i++) { putmc($pos,$i*4); }
putbigmc($pos,$i*4);

$pos++;
for ($i = 1; $i <= $n-$k-1; $i++) {
	for ($j = 1; $j < $k; $j++) { putx($pos,$j*4+$i*2-4); }
	putbigmc($pos,$j*4+$i*2-4);
	$pos += 2;
}

for ($i = 0; $i < $k-1; $i++) { putmc($pos,2*($n-$k)+$i*4); }

# horizontal wall (center left)
put($k,$half,'%'x($vpos-$k));

# central vertical wall
put($vpos,0,('%')x$height);	# vertical wall

# punch holes and put mirrors for each node
for ($i = 0; $i < $n; $i++) {
	put($vpos,$outrow[$i],' ');
	put($vpos,$outrow[$i]+$half,' ');
	put($vpos+1+$i,$outrow[$i],'M');
	put($vpos+1+$i,$outrow[$i]+$half,'M');
}

# looping edges
for $i (@loops) { put($vpos,$outrow[$i],"D"); }

# non-looping edges
for $e (@edges) {
	($a,$b) = split(' ',$e);
	put($vpos+$a+1,$outrow[$b],"D");
}

########### print puzzle to stdout ###########

for $line (@puzzle) { print join("",@$line); }
exit 0;

########### auxilary procedures ###########

sub putmc {
	my ($x,$y) = @_;
	put($x,$y+1,'M','M','M','%');
	put($x,$y+$half+1,'M','M','M','%');
}

sub putbigmc {
	my ($x,$y) = @_;
	put($x,$y+1,'M','M',' ','M','%');
	put($x,$y+$half+1,'M','M',' ','M','%');
}

sub putx {
	my ($x,$y) = @_;
	put($x,$y+1,'MM','MM','MM','%%');
	put($x,$y+$half+1,'MM','MM','MM','%%');
}

sub put {
	my ($x,$y,@l) = @_;
	$y += $baserow;
	for my $s (@l) {
		my $line = $puzzle[$y++];
		my @nc = split(//,$s);
		splice @$line,$x,$#nc+1,@nc;
	}
}
